#include "dict.h"

#include <KCodecs>
#include <KLocale>
#include <QTcpSocket>
#include <QTimer>
#include <QRegExp>


const char *Dict::DefaultHost = "dict.org";
const int Dict::DefaultPort = 2628;
const char *Dict::DefaultDatabase = "!";
const char *Dict::DefaultStrategy = ".";
const char *Dict::AllDatabases = "*";

Dict::Dict(const QString &host, int port)
	: socket(0), state(Disconnected)
{
	dict.host = host;
	dict.port = port;
	setupSocket();
	setupTimers();
}

void Dict::reset()
{
	delete socket;
	state = Disconnected;
	setupSocket();
}

void Dict::abort()
{
	// Qt before 4.5 cannot abort the connection in connecting state
	if (qVersion() < QString("4.5") && state == Connecting)
		reset();
	else
		socket->abort();
}

void Dict::setupSocket()
{
	socket = new QTcpSocket(this);
	connect(socket, SIGNAL(readyRead()), SLOT(read()));
	connect(socket, SIGNAL(stateChanged(QAbstractSocket::SocketState)),
		SLOT(socketState(QAbstractSocket::SocketState)));
	connect(socket, SIGNAL(error(QAbstractSocket::SocketError)),
		SLOT(socketError(QAbstractSocket::SocketError)));
}

void Dict::setupTimers()
{
	setTimeouts();
	connect(&connectTimer, SIGNAL(timeout()), SLOT(connectTimeout()));
	connect(&readTimer, SIGNAL(timeout()), SLOT(readTimeout()));
	connect(&idleTimer, SIGNAL(timeout()), SLOT(idleTimeout()));
}

void Dict::setTimeouts(int connect, int read, int idle)
{
	connectTimer.setInterval(connect);
	readTimer.setInterval(read);
	idleTimer.setInterval(idle);
}

void Dict::connectTimeout()
{
	connectTimer.stop();
	abort();
	request.clear();
	queue.clear();
	emit errorMessage(i18n("Connect Timeout"), ConnectTimeout);
	emit statusMessage(i18n("Canceled"));
}

void Dict::readTimeout()
{
	readTimer.stop();
	abort();
	request.clear();
	queue.clear();
	emit errorMessage(i18n("Read Timeout"), ReadTimeout);
	emit statusMessage(i18n("Canceled"));
}

void Dict::idleTimeout()
{
	send("quit");
}

void Dict::open()
{
	socket->connectToHost(dict.host, dict.port);
	connectTimer.start();
}

void Dict::setHost(const QString &host, int port)
{
	if (dict.host != host || dict.port != port) {
		dict.host = host;
		dict.port = port;
		if (state != Disconnected)
			abort();
		queue.clear();
		dict.ident.clear();
		dict.user.clear();
	}
}

void Dict::read()
{
	readTimer.stop();
	int bytes = socket->bytesAvailable();
	if (!bytes) return;

	if (state == Writing || state == Connected) state = Reading;
	emit statusMessage(i18n("Receiving"));
	QByteArray buf = socket->read(bytes);

	// convert from utf only at newline
	// since a multibyte character can get split across packets
	int end = buf.lastIndexOf('\n');
	read(QString::fromUtf8(extra + buf.left(end)).remove('\r'));
	extra = buf.mid(end + 1);
}

void Dict::send(const QString &command)
{
	if (state > Connected) {
		state = Writing;
		socket->write(command.toUtf8() + '\n');
		readTimer.start();
		idleTimer.start();
		emit statusMessage(i18n("Sending"));
	} else {
		if (state == Disconnected)
			open();
		queue.append(command);
	}
}

void Dict::cancel()
{
	request.clear();
	queue.clear();
	connectTimer.stop();
	readTimer.stop();

	if (state > Connected) {
		state = Canceling;
		idleTimer.start();
	} else if (state != Disconnected)
		abort();

	emit statusMessage(i18n("Canceled"));
}

void Dict::status()
{
	request.type = DictRequest::Status;
	request.command = "status";
	send(request.command);
}

void Dict::define(const QString &query, const QString &database)
{
	Q_ASSERT(!database.isEmpty());

	request.type = DictRequest::Define;
	request.query = query;
	request.database = database;
	request.command = QString("define %1 \"%2\"")
		.arg(database, request.query.remove('\"'));
	send(request.command);
}

void Dict::match(const QString &query, const QString &database, const QString &strategy)
{
	Q_ASSERT(!database.isEmpty());
	Q_ASSERT(!strategy.isEmpty());

	request.type = DictRequest::Match;
	request.query = query;
	request.database = database;
	request.strategy = strategy;
	request.command = QString("match %1 %2 \"%3\"")
		.arg(database, strategy, request.query.remove('\"'));
	send(request.command);
}

void Dict::info(const QString &query)
{
	if (query == "server")
		request.type = DictRequest::Server;
	else if (query == "databases")
		request.type = DictRequest::Databases;
	else if (query == "strategies")
		request.type = DictRequest::Strategies;
	else
		request.type = DictRequest::Info;

	request.command = QString("show %1").arg(query);
	send(request.command);
}

void Dict::setClient(const QString &client)
{
	dict.client = client;
}

void Dict::setUser(const QString &username, const QString &password)
{
	dict.user = username;
	dict.password = password;
}

void Dict::auth(const QString &username, const QString &password)
{
	setUser(username, password);
	if (dict.user.isEmpty())
		return;
	request.type = DictRequest::Auth;
	request.command = "";
	(state >= Connected) ? auth() : open();
}

void Dict::auth()
{
	KMD5 context(dict.msgid.toAscii());
	context.update(dict.password.toAscii());

	send(QString("auth %1 ").arg(dict.user) + context.hexDigest());
}

void Dict::serverIdent(const QString &text)
{
	QRegExp rx("(.*) <(.*)> (<.*>)");
	rx.indexIn(text);
	QStringList list = rx.capturedTexts();
	dict.ident = list[1];
	dict.capabilities = list[2].split('.');
	dict.msgid = list[3];
}

inline DictResponse::ResponseCode responseCode(const QString &line)
{
	return (DictResponse::ResponseCode)
		((line.size() >= 3) ? line.left(3).toInt() : 0);
}

inline QString responseMessage(const QString &line)
{
	return (line.size() > 4) ? line.mid(4) : "";
}

inline bool endMark(const QString &line)
{
	return (line == ".");
}

void Dict::read(const QString &text)
{
	if (state == Reading) state = Parsing;

	foreach(QString line, text.split('\n')) {
		if (state == Parsing) {
			if (!line.isEmpty())
				parse(line);
		} else if (state == Appending) {
			if (endMark(line)) {
				if (response.type == DictResponse::Words)
					response.text += '\n';
				state = Parsing;
			} else
				append(line);
		}
	}

	if (state == Parsing) state = Reading;
}

DictPair wordpair(const QString &line)
{
	QString key = line.section('\"', 2, 2).trimmed();
	QString val = line.section('\"', 3, 3);
	return qMakePair(key, val);
}

DictPair matchpair(const QString &line)
{
	QString key = line.section(' ', 0, 0);
	QString val = line.section('\"', 1, 1);
	return qMakePair(key, val);
}

void Dict::parse(const QString &line)
{
	DictResponse::ResponseCode code = responseCode(line);

	if (code != DictResponse::OK) {
		response.type = code;
		response.message = responseMessage(line);
		emit responseReceived(response);
	}
	switch (code) {
	case DictResponse::Definitions:
		response.text.clear();
		response.database.clear();
		break;
	case DictResponse::Words:
		pair = wordpair(response.message);
		if (pair.first != response.database && !response.database.isEmpty()) {
			emit responseProcessed(response);
			response.text.clear();
		}
		response.database = pair.first;
		response.description = pair.second;
		state = Appending;
		break;
	case DictResponse::Matches:
		response.matches.clear();
		response.database.clear();
		state = Appending;
		break;
	case DictResponse::OK:
		switch (response.type) {
		case DictResponse::Words:
		case DictResponse::Matches:
			response.done = true;
			break;
		case DictResponse::DatabaseList:
		case DictResponse::StrategyList:
		case DictResponse::DatabaseInfo:
		case DictResponse::ServerInfo:
			break;
		case DictResponse::Hello:
		case DictResponse::Bye:
			return;
			break;
		default:
			qWarning() << "unhandled code" << response.type;
			return;
		}
		done();
		response.done = false;
		break;
	case DictResponse::NoMatch:
		response.text = i18n("No matches found");
		done();
		break;
	case DictResponse::DatabaseList:
	case DictResponse::StrategyList:
	case DictResponse::ServerInfo:
	case DictResponse::DatabaseInfo:
		response.list.clear();
		state = Appending;
		break;
	case DictResponse::Hello:
		serverIdent(response.message);
		if (!dict.client.isEmpty())
			send(QString("client \"%1\"").arg(dict.client));
		if (!dict.user.isEmpty())
			auth();
		if (!queue.isEmpty()) {
			foreach(QString command, queue)
				send(command);
			queue.clear();
		}
		break;
	case DictResponse::Bye:
		break;
	case DictResponse::StatusMessage:
		response.text = i18n("Connected to %1.", dict.host);
		done();
		break;
	case DictResponse::AuthOK:
		emit authenticated(true);
		if (request.type == DictRequest::Auth) {
		response.text = i18n("Authenticated to %1.", dict.host);
		done();
		}
		break;
	case DictResponse::AuthDenied:
		emit authenticated(false);
		if (request.type == DictRequest::Auth)
		emit errorMessage(i18n("Authentication failed"), code);
		break;
	case DictResponse::AccessDenied:
		emit errorMessage(i18n("Access denied"), code);
		break;
	case DictResponse::InvalidDatabase:
		emit errorMessage(i18n("Invalid database: %1", request.database), code);
		break;
	case DictResponse::InvalidStrategy:
		emit errorMessage(i18n("Invalid strategy: %1", request.strategy), code);
		break;
	case DictResponse::NoDatabase:
		emit errorMessage(i18n("No databases available"), code);
		break;
	case DictResponse::NoStrategy:
		emit errorMessage(i18n("No strategies available"), code);
		break;
	case DictResponse::ServerUnavailable:
		emit errorMessage(i18n("Server temporarily unavailable"), code);
		break;
	case DictResponse::ServerShutdown:
		emit errorMessage(i18n("Server shutting down"), code);
		break;
	default:
		qWarning() << "unhandled code" << code;
	}
}

void Dict::done()
{
	emit responseProcessed(response);
	emit responseText(response.text);
	emit statusMessage(i18n("Done"));
	response.clear();
}

void Dict::append(const QString &line)
{
	switch (response.type) {
	case DictResponse::Words:
		response.text += line;
		break;
	case DictResponse::Matches:
		if (line.isEmpty()) return;
		pair = matchpair(line);
		if (pair.first != response.database && !response.database.isEmpty()) {
			emit responseProcessed(response);
			response.matches.clear();
		}
		response.database = pair.first;
		response.matches.append(pair.second);
		break;
	case DictResponse::DatabaseList:
		if (line.isEmpty()) return;
		response.list.append(matchpair(line));
		break;
	case DictResponse::StrategyList:
		if (line.isEmpty()) return;
		response.list.append(matchpair(line));
		break;
	case DictResponse::ServerInfo:
		response.text += line;
		break;
	case DictResponse::DatabaseInfo:
		response.text += line;
		break;
	default:
		qWarning() << "unhandled code" << response.type;
	}
	if (!response.isEmpty()) response.text += '\n';
}

void Dict::socketState(QAbstractSocket::SocketState socketState)
{
	QString message;
	switch (socketState) {
	case QAbstractSocket::UnconnectedState:
		readTimer.stop();
		idleTimer.stop();
		state = Disconnected;
		message = i18n("Disconnected");
		emit connected(false);
		break;
	case QAbstractSocket::HostLookupState:
		state = HostLookup;
		message = i18n("Looking for %1...", dict.host);
		break;
	case QAbstractSocket::ConnectingState:
		state = Connecting;
		message = i18n("Connecting to %1...", dict.host);
		break;
	case QAbstractSocket::ConnectedState:
		connectTimer.stop();
		readTimer.start();
		state = Connected;
		message = i18n("Waiting for %1...", dict.host);
		emit connected(true);
		break;
	case QAbstractSocket::ClosingState:
		message = i18n("Disconnecting");
		break;
	default:
		qWarning() << "unhandled state" << socket->state();
	}
	if (!message.isEmpty())
		emit statusMessage(message);
}

void Dict::socketError(QAbstractSocket::SocketError error)
{
	connectTimer.stop();
	QString message;
	NetworkError type = UnknownError;
	switch (error) {
	case QAbstractSocket::RemoteHostClosedError:
		break;
	case QAbstractSocket::ConnectionRefusedError:
		message = i18n("Connection refused by %1", dict.host);
		type = ConnectionRefused;
		break;
	case QAbstractSocket::HostNotFoundError:
		message = i18n("Host %1 not found", dict.host);
		type = HostNotFound;
		break;
	default:
		message = socket->errorString();
// 		qDebug() << "unhandled error" << error;
	}
	if (!message.isEmpty())
		emit errorMessage(message, type);
}
