#include "dictpage.h"

#include <KLocale>
#include <QFile>
#include <QRegExp>


int DictPage::maxMatches = 250;
int DictPage::maxSuggestions = 12;

DictPage::DictPage(const DictResponse &response)
{
	correct = false;
	set(response);
}

void DictPage::set(const DictResponse &response, bool append)
{
	switch (response.type) {
	case DictResponse::Words:
		words(response.text);
		if (headings) heading(response.description);
		break;
	case DictResponse::Matches:
		if (correct) {
			if (!append) corrections(true);
			suggestions(response.database, response.matches);
			append = true;
		} else {
			matches(response.matches);
			if (headings) heading(response.database);
		}
		break;
	case DictResponse::NoMatch:
		return;
		break;
	case DictResponse::DatabaseList:
		databases(response.list);
		break;
	case DictResponse::StrategyList:
		strategies(response.list);
		break;
	case DictResponse::DatabaseInfo:
		words(response.text);
		break;
	case DictResponse::ServerInfo:
		server(response.text);
		break;
	case DictResponse::StatusMessage:
		status(response.message);
		break;
	default:
		html = response.text;
	}
	typeCode = response.type;
	emit changed(append);
}

void DictPage::nomatches(const QString &text, bool define)
{
	query = text;
	html = "<div class=\"message\">" + ((define)
		? i18n("No definitions found for <b>%1</b>.", text)
		: i18n("No matches found for <b>%1</b>.", text))
		+ "</div>";
	typeCode = DictResponse::NoMatch;
	emit changed();
}

void DictPage::corrections(bool found)
{
	if (found)
		html = i18n("<p>Suggestions:");
	else
		html = i18n("<p>No suggestions.");
	emit changed(true);
}

void DictPage::heading(const QString &text)
{
	html.prepend("<div class=\"heading\">" + text + "</div>");
}

void DictPage::words(const QString &text)
{
	html = text;
	// replace html tags
	html.replace('<', "&lt;");
	html.replace('>', "&gt;");
	html.prepend("<div class=\"definition\">");
	
	int start = 0;
	while ((start = html.indexOf('{', start + 1)) >= 0) {
		int end = html.indexOf('}', start);
		// missing matching brace
		if (end < 0) break;

		int len = end - start;
		// too long for a link
		if (len > 240) continue;

		QString link = html.mid(start + 1, len - 1);
		// multiple lines inside braces, could be source code
		if (link.count('\n') > 1) continue;

		int pos;
		// skip external links
		if ((pos = link.indexOf('(')) >= 0 && link.indexOf(')') > pos) {
			html.replace(start, len + 1, link);
			continue;
		}
		// format link
		html.replace(start, len + 1, QString("<a href=\"%1\">%2</a>")
			.arg(link.simplified(), link));
	}
	html.append("</div>");
}

void DictPage::server(const QString &text)
{
	html = "<div class=\"info\">";
	html += text;
	html += "</div>";
}

void DictPage::status(const Dict *dict)
{
	html = "<div class=\"definition\">";
	html += i18n("Connected to dict server\n");
	html += i18n("\nHost: %1", dict->host());
	html += i18n("\nPort: %1", QString::number(dict->port()));
	if (!dict->ident().isEmpty())
		html += i18n("\nIdentification: %1", dict->ident());
	if (!dict->capabilities().isEmpty()) {
		html += i18n("\nCapabilities:");
		foreach(QString text, dict->capabilities())
			html += " " + text;
	}
	if (!dict->client().isEmpty())
		html += i18n("\n\nClient: %1", dict->client());
	if (!dict->user().isEmpty())
		html += i18n("\nUser: %1", dict->user());
	html += "</div>";
}

void DictPage::status(const QString &message)
{
	QRegExp rx("d/m/c = (.*)/(.*)/(.*);");
	html += "<div class=\"definition\">";
	if (rx.indexIn(message) > 0) {
		QStringList list = rx.capturedTexts();
		html += "<div class=\"definition\">";
		html += "\nConnection Status";
		html += "\nDefines: " + list[1];
		html += "\nMatches: " + list[2];
		html += "\nComparisons: " + list[3];
		html += "</div>";
	} else
		html += "\nStatus Message\n" + message;
	html += "</div>";
}

void DictPage::matches(const QStringList &list)
{
	html = "<div class=\"match\">";
	int count = qMin(list.size(), maxMatches);
	for (int i = 0; i < count; ++i)
		html += QString("<a href=\"%1\">%1</a>\n").arg(list[i]);
	html += "</div>";
}

void DictPage::suggestions(const QString &database, const QStringList &list)
{
	html = "<div class=\"list\">";
	if (headings) html += database + ": ";
	html += "<span class=\"suggestion\">";
	int count = qMin(list.size(), maxSuggestions);
	for (int i = 0; i < count; ++i)
		html += QString("<a href=\"%1\">%1</a>, ").arg(list[i]);
	html.chop(2);
	html += "</span></div>";
}

void DictPage::databases(const DictList &list)
{
	html = "<table>";
	foreach (DictPair pair, list) {
		html += QString("<tr><td class=\"name\"><a href=\"about:%1\">%1</a><td>").arg(pair.first);
		html += "<td>" + pair.second;
	}
	html += "</table>";
}

void DictPage::strategies(const DictList &list)
{
	html = "<table>";
	foreach (DictPair pair, list) {
		html += "<tr><td class=\"name\">" + pair.first;
		html += "<td>" + pair.second;
	}
	html += "</table>";
}

void DictPage::about(const QString &name, const QString &version)
{
	QFile file(QString(":intro.html"));
	bool open = file.open(QFile::ReadOnly);
	Q_ASSERT(open);
	html = QString(QString(file.readAll())).arg(name, version);
	typeCode = 0;
	emit changed();
}

void DictPage::error(const QString &text, int code)
{
	QString title, desc;
	QString network = i18n("If you are unable to connect to any host, check your computer's network connection.");
	QString firewall = i18n("If your computer or network is protected by a firewall, check that you can access the network.");
	QString unavail =  i18n("The server could be temporarily unavailable or too busy. Try again in a few moments.");
	QString typing = i18n("Check the address for typing errors.");
	QString web = i18n("If there is a web server on the same address, it may provide information on how to obtain access.");

	switch (code) {
	case Dict::HostNotFound:
		title = i18n("Unable to find the host.");
		desc = "<li>" + typing;
		desc += "<li>" + network + "<li>" + firewall;
		break;
	case Dict::ConnectionRefused:
		title = i18n("Unable to connect to the server.");
		desc = "<li>" + i18n("The host may not run a dict server.");
		desc += "<li>" + typing + "<li>" + unavail;
		break;
	case Dict::ConnectTimeout:
		title = i18n("Unable to connect to the server.");
		desc = "<li>" + unavail;
		desc += "<li>" + network + "<li>" + firewall;
		break;
	case Dict::ReadTimeout:
		title = i18n("Unable to read from the server");
		desc = "<li>" + unavail;
		desc += "<li>" +  i18n("If you receive frequent timeouts, the connection is to slow. Try a different dict server.");
		break;
	case MalformedURL:
		desc = "<li>" + typing;
		break;
	case UnsupportedScheme:
		title = i18n("The protocol is not supported.");
		desc = "<li>" + i18n("Supported protocols are <b>dict</b> and <b>http</b>.");
		break;
	case FileSystem:
		title = i18n("Unable to save the document.");
		desc = "<li>" + i18n("Check that you have the permission to write the file.");
		break;
	case DictResponse::ServerUnavailable:
		desc = "<li>" + i18n("The server is temporarily unavailable or too busy. Try again in a few moments.");
		break;
	case DictResponse::ServerShutdown:
		desc = "<li>" + i18n("The server is shutting down and it is unable to serve any more requests.");
		break;
	case DictResponse::AccessDenied:
		title = i18n("Your computer is not allowed to connect.");
		desc = "<li>" + web;
		break;
	case DictResponse::AuthDenied:
		desc = "<li>" + i18n("See <a href=\"about:server\">Server Info</a> page for information about the server.");
		desc += "<li>" + web;
		break;
	case UnknownDatabase:
	case DictResponse::InvalidDatabase:
		desc = "<li>" + i18n("See <a href=\"about:databases\">Database Info</a> page for a list of available databases.");
		break;
	case DictResponse::InvalidStrategy:
		desc = "<li>" + i18n("See <a href=\"about:strategies\">Strategy Info</a> page for a list of available strategies.");
		break;
	default:
		qDebug() << "unhandled error" << code;
	}

	QFile file(QString(":error.html"));
	bool open = file.open(QFile::ReadOnly);
	Q_ASSERT(open);
	html = QString(QString(file.readAll()))
		.arg(text, title, desc);
	typeCode = code;
	emit changed();
}

void DictPage::warning(const QString &text, Warning type)
{
	QString title, desc;

	switch (type) {
	case LongSelection:
		title = i18n("Long Selection");
		desc = i18n("Selected text is longer than configured limit.");
		break;
	case UrlSelection:
		title = i18n("Url Selected");
		desc = i18n("Selected text is a Url.");
		break;
	}
	QFile file(QString(":warning.html"));
	bool open = file.open(QFile::ReadOnly);
	Q_ASSERT(open);
	html = QString(QString(file.readAll()))
		.arg(title, text, desc);
	typeCode = type;
	emit changed();
}
